/*---------------------------------------------------------------
 $Log: slope.c,v $

 Revision 1.5  2002/02/24 07:35:22  gxd
 Add MaxSessions to process in command line.

 Revision 1.4  2002/02/24 07:11:38  gxd
 Remove dual link support in data structure.


 --------------------------------------------------------*/
#ifndef lint
static char vcid[] = "$Id: slope.c,v 1.5 2002/02/24 07:35:22 gxd Exp gxd $";
#endif /* lint */

#include <stdio.h>
#include <stdlib.h>
#include<string.h>
#include <stdarg.h>
#include <assert.h>
#include <time.h>
#include <math.h>

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/

#define BLKSIZE     256         /* block size for enlarging vectors */
#define REMOVED     NULL        /* return value in DecItem() */

/* #define MAXSESSIONSTOPROCESS  1024*1024 */ /* put this to user specified variable */

typedef struct _node
{
  int iid, support;
  /*  struct _node *prev, *next; */
} NODE;

#define MINBULKSIZE           16
#define BULKS                 8192
#define HashEnlargeThreshold  0.2  /* inc bulksize when these percent units allocated */
#define HashShrinkThreshold   0.01  /* dec bulksize when these percent units allocated */
#define NORMALBULKSIZE        16

typedef struct
{
  int bulksize;
  int cnt;
  NODE *bulk[MINBULKSIZE * BULKS];
} HASH;
#define DELETED ((NODE *)1)
#define IsPointer(bulki) ((bulki)>DELETED)

typedef struct {
  HASH *hash;
  /* NODE *head; */
  int  NumSessions;
  int  TotalCount;  /* Total number of items */
} CF; /* cluster feature */

typedef struct
{
  int size, cnt;
  int ids[1];
} SESSION;

typedef struct {
  int sid;
  int cid;
} SidCid;

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static unsigned long TotalSessions;    /* sesion file format */
static unsigned long TotalItems;       /* first 4 bytes: total sessions */
static struct {                        /* second 4 bytes: total distinct Items */
  unsigned long int sid;               /* then are requests (8 bytes each) */
  unsigned long int iid;               /* strictly ordered by sid */
} request;                             
static char    sessionfilename[256];   /* input sessionfile */
static char    itemfilename[256];      /* the itemfile according to the sessionfile */
static char    resultfilename[256];    /* result of clustering */
static char    clusterfilename[256];   
static char    plotfilename[256];      /* used for gnuplot */
static double  Parameter = 1.2;
/* restrict the number of clusters, or the clustering speed with be low */
static int     MaxClusters = 50; 
static int     RotateTimes = 100;
static long    MaxSessions = 1024*1024;
static int     OutputMode;
static CF     *cf = NULL;              /* pointer to array of cfs, dynamic growth */
static int     cfcnt = 0;
static int     cfsize = 0;
static int    *Result = NULL;          /* pointer to cluster result, dynamic growth */
static int     resultcnt = 0; 
static int     resultsize = 0;
static double  TotalProfitLargestCluster = 0;
static double  TotalProfitSmallestCluster = 0;

/* control the verbose level of debug information */
#define INFO_INFO  1
#define INFO_WARN  2
#define INFO_INFO2 3
#define INFO_DEBUG 4
#define INFO_DEBUG2 5
static int    VerboseLevel=INFO_INFO;



/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

void error(const char *errmsg)  /* -- show error message and exit -- */
{
  fprintf(stderr, "%s\n", errmsg);
  exit(1);
}

void info(int level, char *fmt, ...)
{
  va_list args;
  
  if (level <= VerboseLevel) {
    va_start(args, fmt);
    vfprintf(stdout, fmt, args);
    fflush(stdout);
    va_end(args);
  }
}

/*--------------------------------------------------------------------*/

/* void DumpOneCluster(int i)
{
  NODE *p;

  assert(i<cfcnt);
  printf("C%d: ", i);
  for (p=cf[i].head; p; p=p->next) {
    if (p!=cf[i].head) printf("->");
    printf("%d:%d", p->iid, p->support);
  }
  printf(" (NS: %d TC: %d)\n", cf[i].NumSessions, cf[i].TotalCount);
}
*/

/* void DumpCluster()
{
  int i;
  
  for (i=0; i<cfcnt; i++) {
    DumpOneCluster(i);
  }
}
*/

void DumpHash(const char *msg, HASH *h)
{
  int i, n;
  
  printf(msg);
  printf("(");
  for (i=0, n=h->cnt; i<h->bulksize*BULKS; i++) {
    if (IsPointer(h->bulk[i])) {
      if (n!=h->cnt) printf(", ");
      printf("%d:%d(%d)", i, h->bulk[i]->iid, h->bulk[i]->support);
      n--;
    }
  }
  printf("):%d\n", h->cnt);
}

/*--------------------------------------------------------------------*/

int IsInSession(SESSION *s, int iid)
{
  int i;

  assert(s&&s->cnt>=1);
  for (i=0; i<s->cnt; i++) if (s->ids[i]==iid) return 1;
  return 0;
}

void DumpSession(SESSION *s)
{
  int i;

  printf("s->cnt: %d (", s->cnt);
  for (i=0; i<s->cnt;i++) if (i!=0)printf(", %d", s->ids[i]); else printf("%d", s->ids[i]);
  printf(")\n");
  fflush(stdout);
}

/*----------------------------------------------------------------------
  Internal data structure related core functions
----------------------------------------------------------------------*/

HASH *NewHash(int bulksize)
{
  HASH *h;

  assert(bulksize>=1);
  h = (HASH *)calloc(sizeof(HASH)+(bulksize - MINBULKSIZE)*BULKS*sizeof(NODE *), 1);
  if (!h) error("Not enough memory.");
  assert(!h->bulk[bulksize*BULKS-1]);
  h->bulksize=bulksize;
  return h;
}

void NewCluster()
{
  void *temp;
  int newsize;

  if (!cf) {
    cf = (CF *)malloc(sizeof(CF)*BLKSIZE);
    if (!cf) error("Not enough memory.");
    cfcnt = 0;
    cfsize = BLKSIZE;
  }
  if (cfcnt == cfsize) {
    newsize = cfsize + ((cfsize < BLKSIZE) ? BLKSIZE : (cfsize >> 1));
    assert(newsize > cfsize);
    temp = (CF *)realloc(cf, sizeof(CF)*newsize);
    if (!temp) error("Not enough memory.");
    cf = temp;
    cfsize = newsize;
  }
  assert(cfcnt < cfsize);
  cf[cfcnt].hash = NewHash(MINBULKSIZE);
  /* cf[cfcnt].head = NULL; */
  cf[cfcnt].TotalCount = 0;
  cf[cfcnt].NumSessions = 0;
  cfcnt++;
}

/*
void InsertNode(NODE **head, NODE *n, NODE *potentialprev)
{
  NODE *p;
  assert(head && n);
  p = *head;
  if (!p) {
    *head = n;
    n->prev = n->next = NULL;
  } else {
    if ((p->support>n->support) 
	|| ( p->support==n->support && p->iid>n->iid)) {
      *head = p->prev = n;
      n->next = p;
      n->prev = potentialprev;
    } else {
      while (p->next && (p->next->support < n->support 
			 || (p->next->support == n->support
			     && p->next->iid < n->iid ))) p = p->next;
      if (p->next) { 
	p->next->prev = n;
      }
      n->prev = p;
      n->next = p->next;
      p->next = n;
    }
  }
}
*/

#define computehash(iid, n) ((iid) % (n))

void ResizeHash(HASH **hashp, int newbulksize)
{
  HASH *new;
  HASH *old;
  int i, j, start;

  assert(hashp && *hashp);
  old = *hashp;
  info(INFO_DEBUG2, "Rearrange! %d / %d => %d\n", 
	 old->cnt, old->bulksize*BULKS, newbulksize*BULKS);

  assert(newbulksize*BULKS > old->cnt);
  new = NewHash(newbulksize);

  for (i=0;i<old->bulksize*BULKS;i++) {
    if (IsPointer(old->bulk[i])) {
      start=computehash(old->bulk[i]->iid, BULKS) * new->bulksize;
      for (j=start; j<new->bulksize * BULKS && IsPointer(new->bulk[j]); j++);
      if (j==new->bulksize*BULKS) {
	for (j=0; j<start && IsPointer(new->bulk[j]); j++);
      }
      assert(!new->bulk[j]);
      new->bulk[j]=old->bulk[i];
      new->cnt++;
    }
  }
  assert(new->cnt == old->cnt);
  free(old);
  *hashp=new;
}

void EnlargeHash(HASH **hashp)
{
  int newbulksize;

  assert(hashp && *hashp);
  newbulksize = ((*hashp)->bulksize < NORMALBULKSIZE) ? 
    2 * (*hashp)->bulksize : 1.5 * (*hashp)->bulksize;
  ResizeHash(hashp, newbulksize);
}

void ShrinkHash(HASH **hashp)
{
  int newbulksize;

  assert(hashp && *hashp);
  if ((*hashp)->bulksize <= NORMALBULKSIZE) return;

  newbulksize = (*hashp)->bulksize >> 1;
  ResizeHash(hashp, newbulksize);
}

NODE *AddToHash(HASH **hashp, int iid)
{
  HASH *h;
  int i, start, found;
  int FirstDeleted = -1;

  assert(hashp && *hashp);
  if ((*hashp)->cnt / (double)((*hashp)->bulksize*BULKS) 
      > HashEnlargeThreshold) {
    EnlargeHash(hashp); 
  }
  h=*hashp;
  assert(h->cnt < h->bulksize*BULKS);  /* ensure there are empty cells */

  found = 0;
  start = computehash(iid, BULKS) * h->bulksize;
  for (i=start; i<start+h->bulksize && h->bulk[i]; i++) { 
    /* search in the belonged bulk */
    if (IsPointer(h->bulk[i]) && h->bulk[i]->iid == iid) { found = 1; break; }
    else if (h->bulk[i]==DELETED && FirstDeleted == -1) { FirstDeleted = i; }
  }
  if (i>=start+h->bulksize && !found) {
    for (; i<BULKS * h->bulksize && h->bulk[i]; i++) {
      /* search in the remaining bulk */
      if (IsPointer(h->bulk[i]) && h->bulk[i]->iid == iid) { 
	found = 1; break; 
      }
      else if (h->bulk[i]==DELETED && FirstDeleted == -1) { FirstDeleted = i; }
    }
    if (i>=BULKS*h->bulksize && !found) {
      for (i=0; i<start && h->bulk[i]; i++) {
	/* search in the starting bulk */
	if (IsPointer(h->bulk[i]) && h->bulk[i]->iid == iid) {
	  found = 1; break; 
	}
	else if (h->bulk[i]==DELETED && FirstDeleted == -1) { 
	  FirstDeleted = i; 
	}
      }
      if (i>=start) {
	assert(FirstDeleted != -1); /* Otherwise, no room in hash */
	i = FirstDeleted;
      }
    }
  }

  assert(!IsPointer(h->bulk[i]) || h->bulk[i]->iid == iid);

  if (IsPointer(h->bulk[i]) && h->bulk[i]->iid == iid) { /* found */
    h->bulk[i]->support++;
  } else {
    assert(!IsPointer(h->bulk[i]));
    h->bulk[i]=malloc(sizeof(NODE)); 
    if (!h->bulk[i]) error("No memory");
    h->bulk[i]->iid=iid;
    h->bulk[i]->support=1;
    h->cnt ++;
  }
  return h->bulk[i];
}

int LocateInHash(HASH *h, int iid)
{
  int i, start, found;

  assert(h);
  if (h->cnt==0) return -1; /* can't find anything in an empty hash */
  found = 0;
  start = computehash(iid, BULKS) * h->bulksize;
  for (i=start; i<start+h->bulksize && h->bulk[i]; i++) { 
    /* search in the belonged bulk */
    if (IsPointer(h->bulk[i]) && h->bulk[i]->iid==iid) { found = 1; break; }
  }
  if (i>=start+h->bulksize && !found) {
    for (; i<BULKS * h->bulksize && h->bulk[i]; i++) {
      /* search in the remaining bulk */
      if (IsPointer(h->bulk[i]) && h->bulk[i]->iid==iid) { found = 1; break; }
    }
    if (i>=BULKS*h->bulksize && !found) {
      for (i=0; i<start && h->bulk[i]; i++) {
	/* search in the starting bulk */
	if (IsPointer(h->bulk[i]) && h->bulk[i]->iid==iid) {
	  found = 1; break; 
	}
      }
    }
  }
  if (!found) {
    return (-1);
  }
  assert(found);
  assert(IsPointer(h->bulk[i]));
  return(i);
}

void RemoveFromHash(HASH **hashp, int iid)
{
  HASH *h;
  int i;

  assert(hashp && *hashp);
  h=*hashp;
  assert(h->cnt < h->bulksize*BULKS);  /* ensure there are empty cells */

  i=LocateInHash(h, iid);
  if (i==-1) {
    printf("Error: %d not find in hash.\n", iid);
    DumpHash("Debug dump: ", h);
    /* DumpCluster(); */
    fflush(stdout);
  }
  assert(i!=-1);
  assert(h->bulk[i]);

  if (--h->bulk[i]->support == 0) {
    free(h->bulk[i]);
    h->bulk[i]=DELETED;
    --h->cnt;
    if ( h->cnt / (double)(h->bulksize*BULKS) < HashShrinkThreshold ) {
      ShrinkHash(hashp); 
    }
  }
}

NODE *IncItem(int clusterid, int iid)
{
  NODE *p;

  assert(clusterid<cfcnt);

  p = AddToHash(&(cf[clusterid].hash), iid);

  /* head: removed code segments for removing p */

  cf[clusterid].TotalCount++;
  return p;
}

NODE *DecItem(int clusterid, int iid)
{
  NODE *p;
  int i;

  info(INFO_DEBUG, "Removing %d from cluster %d ...\n", iid, clusterid);
  assert(clusterid<cfcnt);

  i = LocateInHash(cf[clusterid].hash, iid);
  assert(i!=-1);

  p = cf[clusterid].hash->bulk[i];

  assert(p&&p->iid==iid);

  cf[clusterid].TotalCount--;

  if (--p->support>0) {
    return p;
  } else {
    /* free p */
    free(p);
    cf[clusterid].hash->bulk[i]=DELETED;
    cf[clusterid].hash->cnt--;
    return REMOVED;
  }
}

/*--------------------------------------------------------------------*/

void AddSession(int clusterid, SESSION *s)
{
  int i;

  assert(s && s->cnt>0 && cfcnt>clusterid);

  for (i=0; i<s->cnt; i++) {
    IncItem(clusterid, s->ids[i]);
  } /* end of for */
  cf[clusterid].NumSessions++;

  info(INFO_DEBUG, "Adding a session to cluster %d ...\n", clusterid);
  /* if (VerboseLevel >=INFO_DEBUG) DumpOneCluster(clusterid); */
}

void RemoveSession(int clusterid, SESSION *s)
{
  int i;

  assert(s && s->cnt>0 && cfcnt>clusterid);

  for (i=0;i<s->cnt;i++) {
    DecItem(clusterid, s->ids[i]);
  }
  cf[clusterid].NumSessions--;

  info(INFO_DEBUG, "Removing a session from cluster %d ...\n", clusterid);
  /* if (VerboseLevel >=INFO_DEBUG)  DumpOneCluster(clusterid); */

}

double Profit()
{
  int i;
  double c;

  for (i=0, c=0; i<cfcnt; i++) {
    if (cf[i].NumSessions > 0)
      c += cf[i].TotalCount * cf[i].NumSessions / pow(cf[i].hash->cnt, Parameter);
  }

  return c;
}

/* DeltaMode */
#define DELTA_ADD 0
#define DELTA_REMOVE 1
/* profit = S * N / pow(L, k) */

double ComputeDeltaProfit(int DeltaMode, int clusterid, SESSION *s)
{
  int i, j, d=0;
  double OldProfit, NewProfit, DeltaProfit = 0;
  
  if (DeltaMode == DELTA_ADD) 
    info(INFO_DEBUG, "computing DeltaProfit:  ADD a session to cluster %d\n", clusterid); 
  else
    info(INFO_DEBUG, "computing DeltaProfit:  REMOVE a session from cluster %d\n", clusterid);

  if (cf[clusterid].NumSessions==0) { /* add to an empty cluster */
    assert(DeltaMode==DELTA_ADD);
    DeltaProfit = pow(s->cnt, 1-Parameter); /*profit = pow(L, 1-k) when N = 1 */
    info(INFO_DEBUG, "\t empty cluster: s->cnt = %d \t DeltaProfit = %f\n", s->cnt, DeltaProfit);
    return DeltaProfit;
  }

  OldProfit = cf[clusterid].TotalCount * cf[clusterid].NumSessions / 
    pow(cf[clusterid].hash->cnt, Parameter);
  if (DeltaMode==DELTA_ADD) {
    for (i=0; i<s->cnt; i++) {
      j = LocateInHash(cf[clusterid].hash, s->ids[i]);
      if (-1 == j)  d++;  /* this is a new item */
    }
    NewProfit = (cf[clusterid].TotalCount + s->cnt) * (cf[clusterid].NumSessions + 1) / 
      pow(cf[clusterid].hash->cnt + d, Parameter);
    DeltaProfit = NewProfit - OldProfit;
  } else {  /* DeltaMode == DELTA_REMOVE */
    assert(DeltaMode==DELTA_REMOVE);
    assert(cf[clusterid].NumSessions>=1);
    for (i=0; i<s->cnt; i++) {
      j = LocateInHash(cf[clusterid].hash, s->ids[i]);
      if (-1 == j) error("Can't find the item in hash table when computing the deltaprofit.");
      if (1 == cf[clusterid].hash->bulk[j]->support)  d++;
    }
    NewProfit = (cf[clusterid].TotalCount - s->cnt) * (cf[clusterid].NumSessions -1) / 
      pow(cf[clusterid].hash->cnt - d, Parameter);
    DeltaProfit = OldProfit - NewProfit;
  }

  info(INFO_DEBUG, "\t OldProfit = %f  NewProfit = %f  DeltaProfit = %f\n", 
       OldProfit, NewProfit, DeltaProfit);

  return DeltaProfit;
}  /* end of ComputeDeltaProfit func */

/* Result[sid]=clusterid */
/* return 1 if different or new, return 0 if same */
int SetResult(int sid, int clusterid)
{
  int newsize;
  void *temp;

  if (Result && sid < resultcnt) {
    if (Result[sid] != clusterid) {
      Result[sid] = clusterid; 
      return 1;
    }
    return 0;
  } else {
    assert(sid == resultcnt);
    if (resultcnt == resultsize) {
      newsize = resultsize + ((resultsize < BLKSIZE) ? BLKSIZE : (resultsize >> 1));
      assert(newsize > resultsize);
      temp = (int *)realloc(Result, sizeof(int)*newsize);
      if (!temp) error("Not enough memory.");
      Result=temp;
      resultsize=newsize;
    }
    assert(resultcnt < resultsize);
    Result[resultcnt++] = clusterid;
    return 1;
  }
}

int iidcomp(const void *i1, const void *i2)
{
  return *(int *)i1 - *(int *)i2;
}

void CleanSession(SESSION *s)
{
  int i, j;

  if (s->cnt<=1) return;
  qsort(s->ids, s->cnt, sizeof(s->ids[0]), iidcomp);
  for (j=0, i=1; i<s->cnt; i++) {
    if (s->ids[i]>s->ids[j]) {
      if (++j!=i) s->ids[j]=s->ids[i];
    }
  }
  s->cnt=j+1;
}


int groupbycid(const void *v1, const void *v2)
{
  SidCid *t1, *t2;

  t1 = (SidCid *)v1;
  t2 = (SidCid *)v2;
  if (t1->cid > t2->cid)
    return 1;
  else if (t1->cid < t2->cid)
    return -1;
  if (t1->sid > t2->sid)
    return 1;
  else 
    return -1;
}

int orderbysupport(const void *v1, const void *v2)
{
  int *t1, *t2;

  t1 = (int *)v1;
  t2 = (int *)v2;
  if (t1[1] > t2[1])
    return -1;
  else if (t1[1] < t2[1])
    return 1;
  if (t1[0] > t2[0])
    return -1;
  if (t1[0] < t2[0])
    return 1;
  return 0;
}

int orderbyNumSessions(const void *v1, const void *v2)
{
  CF *t1, *t2;

  t1 = (CF *)v1;
  t2 = (CF *)v2;
  if (t1->NumSessions < t2->NumSessions)
    return -1;
  else if (t1->NumSessions > t2->NumSessions)
    return 1;
  return 0;
}

void DumpTopItems()
{
  int i, j, k;
  int *ClusterItems=NULL;
  int item_array_size, item_array_cnt, newsize;
  void *temp;
  char **item_array;
  FILE *itemfile;
  char item_str[2048];

  /* read the item name from itemfile into item_array */
  printf("Reading itemfile ..."); fflush(stdout);
  item_array_size = BLKSIZE;
  item_array_cnt = 0;
  item_array = malloc(sizeof(char *) * item_array_size);
  itemfile = fopen(itemfilename, "r");
  if (!itemfile) error("Error opening item file.");

  /*  while (fscanf(itemfile,"%s\n", item_str) != EOF) {*/
    while (fgets(item_str, sizeof(item_str), itemfile) != NULL) {
    if (item_array_cnt == item_array_size) {
      newsize = item_array_size + (item_array_size >> 1);
      assert(newsize > item_array_size);
      temp = realloc(item_array, sizeof(char *)*newsize);
      if (!temp) error("Not enough memory.");
      item_array = (char **)temp;
      item_array_size = newsize;
    }
    item_array[item_array_cnt] = malloc(strlen(item_str) + 1);
    if (!item_array[item_array_cnt]) error("Not enough memory.");
    if (strlen(item_str)>0 && item_str[strlen(item_str)-1]=='\n') {
      item_str[strlen(item_str)-1]='\0'; /* remove the last CR/LF */
    }
    strcpy(item_array[item_array_cnt],item_str);
    item_array_cnt++;
  }
  fclose(itemfile);
  printf("done.\n");

  qsort(cf, cfcnt, sizeof(CF), orderbyNumSessions);  /* sort cf[] order by NumSessions */

  for (i=0; i<cfcnt; i++) {
    if (cf[i].NumSessions>1) { /* dump detailed item name for non-singleton clusters */
      printf("C%d: N=%d, L=%d, S=%d\n", 
	     i+1, cf[i].NumSessions, cf[i].hash->cnt, cf[i].TotalCount);
      ClusterItems=calloc(sizeof(int), 2*cf[i].hash->cnt);
      if (!ClusterItems) error("Not enough memory.");
      for (j=0,k=0;j<cf[i].hash->bulksize*BULKS;j++) {
	if (IsPointer(cf[i].hash->bulk[j])) {
	  ClusterItems[k++]=cf[i].hash->bulk[j]->iid;
	  ClusterItems[k++]=cf[i].hash->bulk[j]->support;
	  assert(k<=cf[i].hash->cnt*2);
	}
      }
      assert(k==cf[i].hash->cnt*2);
      qsort(ClusterItems, cf[i].hash->cnt, sizeof(int)*2, orderbysupport);
      for (k=0; k<20&&k<2*cf[i].hash->cnt; k+=2) {  /* modify k<20 to show more items */
	printf("\t%s, support=%d\n", 
	       item_array[ClusterItems[k]], ClusterItems[k+1]);
      }
      free(ClusterItems);
    }
  }
  
  /* free item array */
  for (i = 0; i < item_array_cnt; i++)  free(item_array[i]);
  if (item_array)  free(item_array);
}

void SaveResult()
{
  int  i, j, k, c, d;
  FILE *resultfile, *sessionfile, *itemfile, *clusterfile, *plotfile = NULL;
  SidCid *sidcid;
  SESSION **session_array, *s;                /* store all sessions */
  NODE *node;
  char item_str[2048], temp_str[256]; 
  void *temp;
  char **item_array;
  int  item_array_size, item_array_cnt;  
  int  newsize;       /* used to change the array size dynamically */
  int  lastsid;

  if (resultcnt <=0 ) {
    printf("No result written.\n");
    return;
  }

  assert(Result);
  sidcid = (SidCid *)malloc(sizeof(SidCid)*resultcnt);
  if (!sidcid) error("Not enough memory allocating sidcid.");

  clusterfile = fopen(clusterfilename, "w");
  if (!clusterfile) error("Error writing cluster file.");
  printf("write the clustering result to %s ...\n", clusterfilename);
  for (i=0; i<resultcnt&&i<MaxSessions; i++) {
    sidcid[i].sid = i;
    sidcid[i].cid = Result[i];
    fprintf(clusterfile, "%d\n", Result[i]);
  }
  fclose(clusterfile);
  if (0 == OutputMode)    return;  /* clusterfile is the only output file */

  qsort(sidcid, resultcnt, sizeof(SidCid), groupbycid);

  resultfile = fopen(resultfilename, "w");
  if (!resultfile) error("Error writing result file.");

  printf("write the detail clustering to %s ...\n", resultfilename);
  printf("session:%d \t cluster:%d\n", resultcnt, cfcnt);
  /* output the result as C0: s0 s1..... */
  c = -1;   j = 0;
  for (i = 0; i<resultcnt; i++) {
    if (c != sidcid[i].cid) {
      c = sidcid[i].cid;
      if (c != 0) fprintf(resultfile, "-- %d -- \n\n", j);
      j = 0;
      fprintf(resultfile, "C%d: \t", c);
    }
    fprintf(resultfile, "s%d  ", sidcid[i].sid);
    j++;
  }

  fprintf(resultfile, "-- %d --\n\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n", j);

  /* read the item name from itemfile into item_array */
  item_array_size = BLKSIZE;
  item_array_cnt = 0;
  item_array = malloc(sizeof(char *) * item_array_size);
  itemfile = fopen(itemfilename, "r");
  if (!itemfile) error("Error opening item file.");

  while (fscanf(itemfile,"%s\n", item_str) != EOF) {
    if (item_array_cnt == item_array_size) {
      newsize = item_array_size + (item_array_size >> 1);
      assert(newsize > item_array_size);
      temp = realloc(item_array, sizeof(char *)*newsize);
      if (!temp) error("Not enough memory.");
      item_array = (char **)temp;
      item_array_size = newsize;
    }
    item_array[item_array_cnt] = malloc(strlen(item_str) + 1);
    if (!item_array[item_array_cnt]) error("Not enough memory.");
    strcpy(item_array[item_array_cnt],item_str);
    item_array_cnt++;
  }
  fclose(itemfile);

  /* read the session id from sessionfile into session_array */
  sessionfile = fopen(sessionfilename, "rb");
  if (!sessionfile) error("Can't open session file."); 
  
  fread(&TotalSessions, sizeof(TotalSessions), 1, sessionfile);  
  fread(&TotalItems, sizeof(TotalItems), 1, sessionfile);
  if ((TotalItems<=0) || (TotalSessions<=0)) {
    printf("Empty file: sessions %ld, items %ld\n", TotalSessions, TotalItems);
    error("Aborted!");
  }

  s = malloc(sizeof(SESSION));
  if (!s) error ("Not enough memoty.");
  s->cnt = 0;
  s->size = 1;
  session_array = calloc(TotalSessions, sizeof(SESSION *));
  if (!session_array) error("Not enough memory.");

  lastsid = -1;  i =0;
  d = fread(&request, sizeof(long), 2, sessionfile);
  while ((d == 2)&&!feof(sessionfile)&&i<MaxSessions) {
    if (lastsid!=request.sid) { /* new session */
      if (lastsid != -1) {
	CleanSession(s); /* sort and remove dups */
	if (s->cnt>=1) { /* only cluster large sessions */
	  /* add the session s to session_array */
	  session_array[i] = malloc(sizeof(SESSION) + (s->cnt -1)* sizeof(int));
	  if (!session_array[i])  error("Not enough memory.");
	  session_array[i]->size = s->cnt;
	  session_array[i]->cnt = s->cnt;
	  for (j = 0; j < s->cnt; j++)
	    session_array[i]->ids[j] = s->ids[j];
	  i++;	      	     	   
	}
      }
      lastsid=request.sid;
      s->cnt=0;
    }
    /* add iid to s */
    if (s->cnt >= s->size) { /* allow more memory */
      assert(s->cnt == s->size);
      newsize = s->size + ((s->size > BLKSIZE) ? (s->size >> 1) : BLKSIZE);
      assert(newsize > s->size);
      s = realloc(s, sizeof(SESSION) + sizeof(int) * (newsize-1));
      if (!s) error("Not enough memory.");	
      s->size=newsize; 
    }
    s->ids[s->cnt++]=request.iid;
    d = fread(&request, sizeof(long), 2, sessionfile);
  } /* end of while for read session file */
  if (lastsid != -1 && i < MaxSessions) {
    CleanSession(s); /* sort and remove dups */
    if (s->cnt>=1) { /* only cluster large sessions */
      /* add the session s to session_array */
      session_array[i] = malloc(sizeof(SESSION) + (s->cnt -1)* sizeof(int));
      if (!session_array[i])  error("Not enough memory.");
      session_array[i]->size = s->cnt;
      session_array[i]->cnt = s->cnt;
      for (j = 0; j < s->cnt; j++)
	session_array[i]->ids[j] = s->ids[j];
      i++;	      	     	   
    }
  }
  fclose(sessionfile);
  
  i = 0;  c=0;
  while (session_array[i] && i<TotalSessions) {
    TotalProfitSmallestCluster += pow(session_array[i]->cnt, 1-Parameter);
    c += session_array[i]->cnt;
    i++;
  }
  TotalProfitLargestCluster = c * i / pow(item_array_cnt, Parameter);
  printf("Total profit for a session being a cluster is %f\n", TotalProfitSmallestCluster);
  printf("Total profit for all sessions being a cluster is %f\n", TotalProfitLargestCluster);

  /* print the cluster:items to the result file */
  for (i=0; i<cfcnt; i++) {
    if (cf[i].NumSessions>0) {
      fprintf(resultfile,"C%d :  Total sessions = %d  Total items = %d----------------\n",
	      i, cf[i].NumSessions, cf[i].hash->cnt);
      c = 1;   /* used as the counter in plotfile */
      if (2 == OutputMode) {         /* output the plotfile */
	sprintf(temp_str, "%s-%d", plotfilename, i);
	plotfile = fopen(temp_str, "w");
	if (!plotfile)  error("Not enough memory.");
      }
      for (k=0; k<cf[i].hash->bulksize*BULKS;k++){
        if (IsPointer(cf[i].hash->bulk[k])) {
	  node=cf[i].hash->bulk[k];
	  if (node->iid > item_array_cnt) error("Itemfile doesn't match Sessionfile.");
	  fprintf(resultfile,"\t( ");
	  j = 0;
	  while (sidcid[j].cid != i) j++;
	  while (sidcid[j].cid == i) {
	    if (IsInSession(session_array[sidcid[j].sid], node->iid))
	      fprintf(resultfile,"S%d ", sidcid[j].sid); 
	    j++;
	  }	
	  fprintf(resultfile, ")\titem=%d : sup=%d  %s\n", node->iid, 
		  node->support,item_array[node->iid]);
	  if (2 == OutputMode) {
	    fprintf(plotfile, "%d\t%d\t%d\t%d\t%s\n", i, c, node->support, 
		    node->iid, item_array[node->iid]); /* clusterid, #, sup, itemID, itemname */
	  }
	  ++c;
	}
      }
      if (2 == OutputMode)  fclose(plotfile);
    }
  }
  fclose(resultfile);

  /* free the space */
  for (i = 0; i < item_array_cnt; i++)  free(item_array[i]);
  if (item_array)  free(item_array);
  for (i = 0; session_array[i] != NULL && i<TotalSessions; i++)  free(session_array[i]);
  if (session_array)  free(session_array);
  free(s);

  free(sidcid);

} /* end of SaveResult */

void PrintPercentageBar(char *c, int percent, int total)
{
  int i;

  assert(total>0);
  for (i=0; i<percent*60/total; i++) printf(c);
}

void DumpClusterResult()
{
  int i, Sum, MaxSessions;

  for (i=0, MaxSessions=Sum=0; i<cfcnt; i++) {
    Sum+=cf[i].NumSessions;
    MaxSessions = MaxSessions>cf[i].NumSessions ?
      MaxSessions : cf[i].NumSessions;
  }

  for (i=0; i<cfcnt; i++) {
    if (cf[i].hash) {
      printf("%2d|", i);
      PrintPercentageBar("-",cf[i].NumSessions, MaxSessions);
      printf("(%d/%d %0.2f%%)\n", 
	     cf[i].NumSessions, Sum, cf[i].NumSessions*100/(float)Sum);
    }
  }
}

int ProcessOneSession(SESSION *s, int pass, int sid)
{
  double MaxDelta, OriginalDelta, d, OldProfit;
  int BestClusterID, i;

  info(INFO_INFO2, "---- Session %d: ", sid);
  if (VerboseLevel>=INFO_INFO2) DumpSession(s); 

  if (cfcnt == 0 || (cf[cfcnt-1].NumSessions!=0 && cfcnt < MaxClusters)) {
    NewCluster(); 
  }  
  
  BestClusterID = -1;
  MaxDelta = OriginalDelta = 0;
  OldProfit=Profit();

  for (i=0; i<cfcnt; i++) {
    info(INFO_DEBUG, "Testing cluster %d ...\n", i);
    if (pass>0 && i == Result[sid]) {
      d = ComputeDeltaProfit(DELTA_REMOVE, i, s);
      OriginalDelta = d;
    } else {
      d = ComputeDeltaProfit(DELTA_ADD, i, s);         /* comput the Profit change */
    }
    if (VerboseLevel==INFO_WARN) printf("*%f ", d);
    if (BestClusterID==-1 || MaxDelta < d ) {
      MaxDelta = d;
      BestClusterID = i;
    }
  }
  assert(BestClusterID!=-1);
  info(INFO_DEBUG, "MaxDelta=%f  BestClusterID=%d\n", MaxDelta, BestClusterID);
  if (pass==0) {
    AddSession(BestClusterID, s);
    info(INFO_WARN, "| %d => %d[%d]\n", 
	 sid, BestClusterID, cf[BestClusterID].NumSessions);
  } else {
    if (MaxDelta > OriginalDelta) {
      /* s should be removed from result[sid], then be added into bestclusterid */
      RemoveSession(Result[sid], s);  
      AddSession(BestClusterID, s);
      info(INFO_WARN, "| %d: %d=>%d[%d] (%d->%d,%d,%d)\n", 
	   sid, Result[sid], BestClusterID, cf[BestClusterID].NumSessions, 
	   OldProfit, Profit(), MaxDelta, OriginalDelta);
    } else {
      /* the original cluster is the best choice, there isn't any change to be done */
      BestClusterID = Result[sid];
      info(INFO_WARN, "| %d: %d[%d] unchanged\n", 
	   sid, Result[sid], cf[Result[sid]].NumSessions);
    }
  }
  return(SetResult(sid, BestClusterID));
}

void ProcessSessionFile()
{
  long    requestsread, sessionsread;
  SESSION *s;
  void    *temp; int newsize;
  int     d, i, k;
  int     lastsid;
  FILE    *sessionfile;
  int     pass, changed;
  double  OldProfit=-1;

  sessionfile = fopen(sessionfilename, "rb");
  if (!sessionfile) error("Can't open session file."); 
  
  fread(&TotalSessions, sizeof(TotalSessions), 1, sessionfile);  
  fread(&TotalItems, sizeof(TotalItems), 1, sessionfile);
  if ((TotalItems<=0) || (TotalSessions<=0)) {
    printf("Empty file: sessions %ld, items %ld\n", TotalSessions, TotalItems);
    error("Aborted!");
  }
  
  s = (SESSION *)malloc(sizeof(SESSION));
  if (!s) error("Not enough memory.");
  s->cnt=0; s->size=1;

  pass = 0; 
  while ((pass == 0 || Profit() > (OldProfit + 0.1))&& (pass < RotateTimes)) {
    if (pass>0) OldProfit = Profit();

    fseek(sessionfile, sizeof(long)*2, SEEK_SET);
    info(INFO_INFO, "%ld sessions with %ld unique items expected. "
	 "Reading ('.' = 1K sessions)\n", 
	 TotalSessions, TotalItems); 
    requestsread=0; sessionsread=0;

    changed = 0;  lastsid = -1;
    d = fread(&request, sizeof(long), 2, sessionfile);
    while ((d == 2)&&!feof(sessionfile)&&sessionsread<MaxSessions) {
      ++requestsread;
      if (lastsid!=request.sid) {    /* new session */
	if (lastsid != -1) {
	  CleanSession(s);        /* sort and remove dups */
	  if (s->cnt>=1) {        /* only cluster large sessions */
	    if (ProcessOneSession(s, pass, sessionsread)) ++changed;
	    if (++sessionsread % (1024) == 0) { 
	      printf("."); 
	      fflush(stdout);
	    }
	  }
	}
	lastsid=request.sid;
	s->cnt=0;
      }
      /* add iid to s */
      if (s->cnt >= s->size) {                /* allocate more memory */
	assert(s->cnt == s->size);
	newsize = s->size + ((s->size > BLKSIZE) ? (s->size >> 1) : BLKSIZE);
	assert(newsize > s->size);
	temp = realloc(s, sizeof(SESSION) + sizeof(int) * (newsize-1));
	if (!temp) error("Not enough memory.");
	s = temp;
	s->size=newsize; 
      }
      s->ids[s->cnt++]=request.iid;
      d = fread(&request, sizeof(long), 2, sessionfile);
    }          /* end of while for read session file */

    if (lastsid != -1 && sessionsread < MaxSessions) {
      CleanSession(s);      /* sort and remove dups */
      if (s->cnt>=1) {      /* only cluster large sessions */
	if (ProcessOneSession(s, pass, sessionsread)) ++changed;
	if (++sessionsread % (1024) == 0) { 
	  printf("."); 
	  fflush(stdout);
	}
      }
    }

    if (sessionsread != TotalSessions && sessionsread < MaxSessions) 
      printf("Warning: sessionsread(%ld) != expected sessions(%ld).\n",
	     sessionsread, TotalSessions);
    ++pass;
    info(INFO_INFO, "\n=== Pass: %d Changed: %d OldProfit: %f Profit: %f ===\n", 
	 pass, changed, OldProfit, Profit());
  }       /* end of while for pass */
  
  fclose(sessionfile);
  free(s);
  DumpClusterResult();
  DumpTopItems();
  SaveResult();

  printf("pass= %d\n",pass);

  /* free the allocated space */
  if (Result)  free(Result);
  for(i = 0; i < cfcnt; i++) {
    for (k=0; k<cf[i].hash->bulksize*BULKS; k++) {
      if (IsPointer(cf[i].hash->bulk[k])) free(cf[i].hash->bulk[k]);
    }
    if (cf[i].hash)  free(cf[i].hash);
  }
  if (cf)  free(cf);

} 
void PrintUsage() 
{
  printf("\n");
  printf("Usage: ./clope DataFile OutputMode [ Parameter [ MaxClusters [RotateTimes [MaxSessions [VerboseLevel]]]]]\n");
  printf("DataFile is the prefix of sessionfile, itemfile and resultfile\n");
  printf("          sessionfile -- DataFile.session\n");
  printf("          itemfile    -- DataFile.item\n");
  printf("          outputfile  -- DataFile.clope_cluster...\n");
  printf("                         DataFile.clope_result...\n");
  printf("                         DataFile.clope_plot...\n");
  printf("OutputMode can be 0, 1, and 2\n");
  printf("          0 -- output cluster file\n");
  printf("          1 -- output cluster file, and result file\n");
  printf("          2 -- output cluster file, result file, and plot file\n");
  printf("Parameter is the adjust factor k (default: 1.2)\n");
  printf("MaxClusters is the maximum cluster number (default: 50)\n");
  printf("RotateTimes is the maximal number of iterations (default: 100)\n");
  printf("MaxSessions is the maximal number of sessions to process (default: 1M)\n");
  printf("\n");
}
/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

int main(int argc, char *argv[])
{
  printf("Clope-based Cluster Algorithm.\n%s\n", vcid);
  if ((argc < 3) || (argc > 8)) {                        /* parameter count is not right */
    PrintUsage();
    exit(1);
  }
  if (argc > 7)  VerboseLevel = atoi(argv[7]);
  if (argc > 6)  MaxSessions = atoi(argv[6]);
  if (argc > 5)  RotateTimes = atoi(argv[5]);
  if (argc > 4)  MaxClusters = atoi(argv[4]);
  if (argc > 3)  Parameter = atof(argv[3]);
  OutputMode = atoi(argv[2]);
  if ( (0 != OutputMode) && (1 != OutputMode) && (2 != OutputMode)) {
    printf("OutputMode value %d is not valid.\n", OutputMode);
    PrintUsage();
    exit(1);
  }
  sprintf(sessionfilename, "%s.session", argv[1]);
  sprintf(itemfilename, "%s.item", argv[1]);
  sprintf(resultfilename, "%s.clope_result.%.3f", argv[1], Parameter);
  sprintf(clusterfilename, "%s.clope_cluster.%.3f", argv[1], Parameter);
  sprintf(plotfilename, "%s.clope_plot.%.3f", argv[1], Parameter);    

  info(INFO_INFO, "Session file: %s\n", sessionfilename);
  info(INFO_INFO, "Item file:    %s\n", itemfilename);
  info(INFO_INFO, "Cluster file: %s\n", clusterfilename);
  info(INFO_INFO, "Result file:  %s\n", resultfilename);
  info(INFO_INFO, "OutputMode:   %d\n", OutputMode);
  info(INFO_INFO, "Parameter:    %.3f\n", Parameter);
  info(INFO_INFO, "MaxCluster:   %d\n", MaxClusters);
  info(INFO_INFO, "RotateTimes:  %d\n", RotateTimes);
  info(INFO_INFO, "MaxSessions:  %d\n", MaxSessions);
  info(INFO_INFO, "VerboseLevel: %d\n", VerboseLevel);

  ProcessSessionFile();
  
  return 0;
}
